#pragma once

#include <utility>
#include <algorithm>
#include <Qt>
#include <QDebug>
#include <QObject>
#include <QThread>
#include <QMutex>
#include <QElapsedTimer>
#include <QWaitCondition>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_drawer.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "owned_qimage.hpp"
#include "_separate_threaded_renderer.hpp"

namespace BGTU::ComputerGraphicsLabWork::QtUtilities {
	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	class SeparateThreadedRenderer : public _SeparateThreadedRenderer_Signals {
	private:
		struct data_t {
		public:
			RendererApi::Sprite<sprite_data_t> *const *sprites;
			std::size_t sprites_count;
			sprite_data_t const *sprite_data;
			std::uint_fast64_t ms_per_frame;
			RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, height;
			RendererApi::Projection *projection;
			RendererApi::Color bg;
		};

		class RendererThread : public QThread {

		private:
			SeparateThreadedRenderer *owner;
			QWaitCondition intr;
		public:

			explicit inline RendererThread(SeparateThreadedRenderer *owner) : QThread{owner}, owner{owner}, intr{} {
			};

			void requestInterruption() {
				QThread::requestInterruption();
				this->intr.wakeAll();
			}

			void force_redraw() {
				this->intr.wakeAll();
			}

		protected:
			void run() final;

		private:
			void sleepms(std::uint_fast64_t ms) {
				QMutex m{};
				m.lock();
				this->intr.wait(&m, ms);
				m.unlock();
			}
		};


		volatile data_t next_data;
		QMutex sync;
		RendererThread *thread;
		QImage *sentinel_img;
		renderer_context_t *renderer_context;
	public:
		explicit SeparateThreadedRenderer(renderer_context_t *renderer_context, QObject *owner = nullptr);

	public slots:

		void set_frame_size(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height) final;

		void set_ms_per_frame(std::uint_fast64_t ms) final;

		void set_sprites(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count);

		void set_sprite_data(sprite_data_t const *data);

		void set_sprites_and_data(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count, sprite_data_t const *data);

		void set_background(RendererApi::Color) final;

	public:
		~SeparateThreadedRenderer();
	};


	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::SeparateThreadedRenderer(renderer_context_t *renderer_context, QObject *owner):
			_SeparateThreadedRenderer_Signals{owner},
			sync{},
			renderer_context{renderer_context},
			sentinel_img{new QImage(0, 0, QImage::Format_ARGB32)},
			thread{new RendererThread(this)},
			next_data{
					.sprites = nullptr,
					.sprites_count = 0,
					.sprite_data = nullptr,
					.ms_per_frame = 1000 / 60,
					.width = 0,
					.height = 0,
					.projection = nullptr
			} {
		this->thread->start();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_frame_size(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->width = width;
		const_cast<data_t *>(&this->next_data)->height = height;
		this->thread->force_redraw();
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_ms_per_frame(std::uint_fast64_t ms) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->ms_per_frame = ms;
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprites(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->sprites = sprites;
		const_cast<data_t *>(&this->next_data)->sprites_count = sprites_count;
		this->thread->force_redraw();
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprite_data(sprite_data_t const *sprite_data) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->sprite_data = sprite_data;
		this->thread->force_redraw();
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_sprites_and_data(RendererApi::Sprite<sprite_data_t> *const *sprites, std::size_t sprites_count, sprite_data_t const *sprite_data) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->sprites = sprites;
		const_cast<data_t *>(&this->next_data)->sprites_count = sprites_count;
		const_cast<data_t *>(&this->next_data)->sprite_data = sprite_data;
		this->thread->force_redraw();
		this->sync.unlock();
	}


	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::set_background(RendererApi::Color c) {
		this->sync.lock();
		const_cast<data_t *>(&this->next_data)->bg = c;
		this->thread->force_redraw();
		this->sync.unlock();
	}

	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	void SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::RendererThread::run() {
		try {
			QElapsedTimer timer;
			while (!this->isInterruptionRequested()) {
				timer.start();
				this->owner->sync.lock();
				data_t cached = *const_cast<data_t *>(&this->owner->next_data);
				this->owner->sync.unlock();

				struct VoxelDrawerExporterImpl : public RendererApi::VoxelDrawer::Exporter {
				public:
					RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t const w;
					RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t const h;
					std::uint_least8_t *const pixels_argb;;
					RendererApi::SpriteMetadata const **const owners;

					VoxelDrawerExporterImpl(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t w, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t h) :
							w{w}, h{h},
							pixels_argb{new std::uint_least8_t[w * h * 4]},
							owners{new RendererApi::SpriteMetadata const *[w * h]} {}

					void set_pixel(long long x, long long y, BGTU::ComputerGraphicsLabWork::RendererApi::Color c, const BGTU::ComputerGraphicsLabWork::RendererApi::SpriteMetadata *owner) final {
						if (x < 0 || this->w <= x) return;
						if (y < 0 || this->h <= y) return;
						this->pixels_argb[(y * this->w + x) * 4] = c.red;
						this->pixels_argb[(y * this->w + x) * 4 + 1] = c.green;
						this->pixels_argb[(y * this->w + x) * 4 + 2] = c.blue;
						this->owners[y * this->w + x] = owner;
					}
				} img{cached.width, cached.height};

				(this->owner->renderer_context->*renderer_procedure)(
						cached.width, cached.height,
						&img,
						cached.bg,
						cached.projection,
						cached.sprites, cached.sprites_count,
						cached.sprite_data
				);
				emit this->owner->frame_rendered(new OwnedQImage{img.pixels_argb, img.w, img.h, QImage::Format_RGB32, img.owners, OwnedQImage::DelegateBuffers});
				auto elapsed = timer.elapsed();
				qCritical() << elapsed << "ms";
				this->sleepms(std::max<std::int_fast64_t>(static_cast<std::int_fast64_t>(cached.ms_per_frame) - elapsed, 0));
			}
		} catch (std::exception const &e) {
			qCritical() << typeid(e).name() << ": " << e.what() << "\n" << "Renderer thread died" << "\n";
		}
	}


	template<class sprite_data_t, class renderer_context_t, _SeparateThreadedRenderer_RenderProcedureType<sprite_data_t, renderer_context_t> renderer_procedure>
	SeparateThreadedRenderer<sprite_data_t, renderer_context_t, renderer_procedure>::~SeparateThreadedRenderer() {
		this->thread->requestInterruption();
		this->thread->wait();
	}
}