#pragma once

#include <cstddef>
#include <cstdlib>
#include <limits>
#include <new>
#include <vector>
#include <algorithm>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_drawer.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "color.hpp"
#include "matrix.hpp"

class QObject;

namespace BGTU::ComputerGraphicsLabWork {
	namespace Utilities {
		class DefaultVoxelDrawerCache {
		private:
			class ZElement {
			public:
				ZElement *next;
				double const z;
				RendererApi::Color::Transparent const color;
				RendererApi::SpriteMetadata *const owner;

				ZElement(ZElement *next, double z, RendererApi::Color::Transparent color, RendererApi::SpriteMetadata *owner) : next{next}, z{z}, color{color}, owner{owner} {}

				void link_after(ZElement *new_next) {
					new_next->next = this->next;
					this->next = new_next;
				}
			};

			class ZElementAllocationBuffer {
			private:
				union Cell {
					ZElement allocated;
					struct Empty {
						Cell *next_empty;
					} empty;
				};


				class Page {
				public:
					Page *next;
				private:

					std::size_t initialized;
					std::size_t capacity;

					Cell buffer[0];

				public:
					explicit Page(Page *next, std::size_t capacity) : next{next}, initialized{0}, capacity{capacity}, buffer{} {};

					Cell *try_allocate();

					static std::size_t calc_capacity(std::size_t bytes) {
						return (bytes - offsetof(Page, buffer)) / sizeof(Cell);
					};

					void reset_allocations() {
						this->initialized = 0;
					}
				};

				Cell *next_unallocated;
				Page *last_page;
				Page sentinel;

			public:
				ZElementAllocationBuffer() : next_unallocated{nullptr}, sentinel{nullptr, 0}, last_page{&this->sentinel} {};

			private:
				Page *_native_extend();

			public:
				template<class ...args_t>
				ZElement *alloc_elem(args_t...args);

				void free_elem(ZElement *e);

			private:
				void _native_free_page(Page *p);

			public:
				void release_resources();

				void reset_allocations();

				~ZElementAllocationBuffer();
			};

			struct pixel_trace_metadata {
			public:
				pixel_trace_metadata() : nearest_z{std::numeric_limits<double>::max()}, voxels{nullptr} {};

				double nearest_z;
				ZElement *voxels;
			};

			class VoxelDrawerImpl : public RendererApi::VoxelDrawer {
			public:
				RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t _width;
				RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t _height;

				pixel_trace_metadata *pixels_metadata;

				ZElementAllocationBuffer *pixel_trace_elements_allocator;
				RendererApi::SpriteMetadata *current_artist;

				VoxelDrawerImpl(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height, pixel_trace_metadata *pixels_metadata, ZElementAllocationBuffer *pixel_trace_elements_allocator);

				[[nodiscard]] inline RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t width() const final {
					return this->_width;
				}

				[[nodiscard]] inline RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t height() const final {
					return this->_height;
				}

				inline pixel_trace_metadata *at(std::uint_fast16_t x, std::uint_fast16_t y) {
					return &(this->pixels_metadata[y * this->_width + x]);
				}

				void add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c) final;
			};


			struct {
				pixel_trace_metadata *buffer;
				std::size_t size;
			} pixel_metadata_cache;
			ZElementAllocationBuffer pixel_trace_elements_allocator;

		public:
			DefaultVoxelDrawerCache();

		private:
			static ZElement *sort_zelements_desc(DefaultVoxelDrawerCache::ZElement *start) noexcept;

		public:
			template<class sprite_data_t>
			void render(
					RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t w,
					RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t h,
					RendererApi::VoxelDrawer::Exporter *frame,
					RendererApi::Color bg,
					RendererApi::Projection const *projection,
					RendererApi::Sprite<sprite_data_t> *const *sprites,
					std::size_t sprites_count,
					sprite_data_t const *sprite_data
			);

			~DefaultVoxelDrawerCache();
		};

		template<class... args_t>
		DefaultVoxelDrawerCache::ZElement *DefaultVoxelDrawerCache::ZElementAllocationBuffer::alloc_elem(args_t... args) {
			auto cell = this->next_unallocated;
			if (cell != nullptr) {
				this->next_unallocated = cell->empty.next_empty;
				return &(cell->allocated);
			}

			while ((cell = this->last_page->try_allocate()) == nullptr)
				this->_native_extend();

			return new(&(cell->allocated)) ZElement{args...};
		}


		template<class sprite_data_t>
		void DefaultVoxelDrawerCache::render(
				RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t w,
				RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t h,
				RendererApi::VoxelDrawer::Exporter *frame,
				RendererApi::Color bg,
				RendererApi::Projection const *projection,
				RendererApi::Sprite<sprite_data_t> *const *sprites,
				std::size_t sprites_count,
				sprite_data_t const *sprite_data
		) {
			if (w == 0 || h == 0)
				return;

			if (this->pixel_metadata_cache.buffer == nullptr || this->pixel_metadata_cache.size < w * h) {
				this->pixel_metadata_cache.buffer = static_cast<pixel_trace_metadata *>(std::realloc(this->pixel_metadata_cache.buffer, w * h * sizeof(pixel_trace_metadata)));
				this->pixel_metadata_cache.size = w * h;
			}

			this->pixel_trace_elements_allocator.reset_allocations();
			VoxelDrawerImpl painter{w, h, this->pixel_metadata_cache.buffer, &this->pixel_trace_elements_allocator};

			for (std::int_fast64_t i = w * h - 1; i >= 0; i--) {
				new(&this->pixel_metadata_cache.buffer[i]) pixel_trace_metadata{};
			}

			for (; sprites_count-- > 0; sprites++) {
				painter.current_artist = *sprites;
				(*sprites)->draw(&painter, projection, sprite_data);
			}

			std::uint_fast16_t y = 0;
			std::uint_fast16_t x = 0;

			for (std::int_fast64_t i = w * h - 1; i >= 0; i--) {
				this->pixel_metadata_cache.buffer[i].voxels = sort_zelements_desc(this->pixel_metadata_cache.buffer[i].voxels);
				RendererApi::Color cc = bg;
				RendererApi::SpriteMetadata *owner = nullptr;
				for (auto p = this->pixel_metadata_cache.buffer[i].voxels; p != nullptr; p = p->next) {
					cc = apply_transparent_color(cc, p->color);
					owner = p->owner;
				}
				frame->set_pixel(x, y, cc, owner);

				if (++x >= w) {
					y++;
					x = 0;
				}
			}
		}
	}

	namespace QtUtilities {
		template<class sprite_data_t, class renderer_context_t, void (renderer_context_t::*renderer_procedure)(RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::visible_pixel_coordinate_fast_t, RendererApi::VoxelDrawer::Exporter *, RendererApi::Color, RendererApi::Projection const *, RendererApi::Sprite<sprite_data_t> *const *, std::size_t, sprite_data_t const *)>
		class SeparateThreadedRenderer;


		template<class sprite_data_t>
		class SeparateThreadedDefaultRendererLinear : public SeparateThreadedRenderer<sprite_data_t, Utilities::DefaultVoxelDrawerCache, &Utilities::DefaultVoxelDrawerCache::render<sprite_data_t>> {
		private:
			Utilities::DefaultVoxelDrawerCache context;
		public:
			explicit SeparateThreadedDefaultRendererLinear(QObject *owner = nullptr) : SeparateThreadedRenderer<sprite_data_t, Utilities::DefaultVoxelDrawerCache, &Utilities::DefaultVoxelDrawerCache::render<sprite_data_t>>{&this->context, owner} {};
		};
	}
}